# -*- coding: utf-8 -*-
import os
import unittest

from skoolkittest import SkoolKitTestCase

TMP_BIN = 'temp.bin'
TMP_SNA = 'temp.sna'
TMP_CTL = 'temp.ctl'
TMP_SFT = 'temp.sft'
TEST_BIN = 'test-sna.bin'
TEST_CTL_C = 'test-sna-c.ctl'
TEST_CTL_G = """c 16384
c 16385
z 16389
c 32768
c 32772
c 32774
c 32777
c 32786
c 32794
z 32801
c 32811
z 32813"""
TEST_CTL_G_HEX = """c $4000
c $4001
z $4005
c $8000
c $8004
c $8006
c $8009
c $8012
c $801A
z $8021
c $802B
z $802D"""

USAGE = """Usage: sna2skool.py [options] file

  Convert a binary (raw memory) file or a 48K SNA or Z80 snapshot into a skool
  file.

Options:
  -c FILE  Use FILE as the control file (default is file.ctl)
  -T FILE  Use FILE as the skool file template (default is file.sft)
  -g FILE  Generate a control file in FILE
  -h       Write hexadecimal addresses in the generated control file
  -H       Write hexadecimal addresses and operands in the disassembly
  -L       Write the disassembly in lower case
  -s ADDR  Specify the address at which to start disassembling (default=16384)
  -o ADDR  Specify the origin address of file.bin (default: 65536 - length)
  -t       Show ASCII text in the comment fields
  -r       Suppress addition of comments listing entry point referrers
  -n N     Set the max number of bytes per DEFB statement to N (default=8)
  -m M     Group DEFB blocks by addresses that divide M
  -z       Write bytes with leading zeroes in DEFB statements
  -l L     Set the max number of characters per DEFM statement to L (default=66)"""

class OptionsTest(SkoolKitTestCase):
    def remove_files(self):
        for f in (TMP_BIN, TMP_SNA, TMP_CTL, TMP_SFT):
            if os.path.isfile(f):
                os.remove(f)

    def _write_text(self, fname, lines):
        f = open(fname, 'wt')
        for line in lines:
            f.write('{0}\n'.format(line))
        f.close()

    def _write_data(self, fname, data):
        f = open(fname, 'wb')
        f.write(bytearray(data))
        f.close()

    def _write_bin(self, data):
        self._write_data(TMP_BIN, data)

    def _write_ctl(self, ctls):
        self._write_text(TMP_CTL, ctls)

    def test_invalid_option(self):
        output, error = self.run_sna2skool('-x {0}'.format(TMP_BIN), err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_no_arguments(self):
        output, error = self.run_sna2skool(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'))

    def test_no_options(self):
        lines = self._write_skool(TEST_BIN, 10)
        self.assertEqual(lines[0], '; Routine at 16384')
        self.assertEqual(lines[1][:10], 'c16384 RET')

    def test_nonexistent_files(self):
        error_tp = 'ERROR: {0}: file not found'

        output, error = self.run_sna2skool(TMP_BIN, err_lines=True)
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0], error_tp.format(TMP_BIN))

        output, error = self.run_sna2skool('-c {0}'.format(TMP_CTL), err_lines=True)
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0], error_tp.format(TMP_CTL))

        tmp_sft = 'temp.sft'
        output, error = self.run_sna2skool('-T {0}'.format(tmp_sft), err_lines=True)
        self.assertEqual(len(error), 1)
        self.assertEqual(error[0], error_tp.format(tmp_sft))

    def test_option_H(self):
        data = [62, 254]           # $FFFB LD A,$FE
        data.extend((195, 1, 128)) # $FFFD JP $8001
        self._write_bin(data)
        self._write_ctl(['c $FFFB'])
        lines = self._write_skool('-H {0}'.format(TMP_BIN), 3)
        self.assertEqual(lines[1][:15], 'c$FFFB LD A,$FE')
        self.assertEqual(lines[2][:15], ' $FFFD JP $8001')

    def test_option_L(self):
        data = [0]      # 65534 nop
        data.append(65) # 65535 defm "A"
        self._write_bin(data)
        self._write_ctl(['c 65534', 'T 65535'])
        lines = self._write_skool('-L {0}'.format(TMP_BIN), 3)
        self.assertEqual(lines[1][:10], 'c65534 nop')
        self.assertEqual(lines[2][:15], ' 65535 defm "A"')

    def test_option_T(self):
        sft = ['; Do stuff']
        sft.append('cC32768,1;20 Blah')
        sft.append(' C32769,3;20 Yah')
        self._write_text(TMP_SFT, sft)
        lines = self._write_skool('-T {0} {1}'.format(TMP_SFT, TEST_BIN), 2)
        self.assertEqual(len(lines), 3)
        self.assertEqual(lines[0], sft[0])
        self.assertEqual(lines[1], 'c32768 LD A,B       ; Blah')
        self.assertEqual(lines[2], ' 32769 JP 32774     ; Yah')

    def test_option_c(self):
        lines = self._write_skool('-c {0} {1}'.format(TEST_CTL_C, TEST_BIN), 10)
        self.assertEqual(self._get_first_address(lines), '32768')

    def test_option_g(self):
        self._write_skool('-g {0} {1}'.format(TMP_CTL, TEST_BIN))
        f = open(TMP_CTL, 'r')
        lines = [line.rstrip() for line in f]
        f.close()
        self.assert_output_equal(TEST_CTL_G.split('\n'), lines)

    def test_option_h(self):
        self._write_skool('-g {0} -h {1}'.format(TMP_CTL, TEST_BIN))
        f = open(TMP_CTL, 'r')
        lines = [line.rstrip() for line in f]
        f.close()
        self.assert_output_equal(TEST_CTL_G_HEX.split('\n'), lines)

    def test_option_l(self):
        bin_len = 16
        self._write_bin([65] * bin_len)
        ctls = ['t {0}'.format(65536 - bin_len)]
        self._write_ctl(ctls)
        defm_len = 5
        defm = 'DEFM "{0}"'.format('A' * defm_len)
        lines = self._write_skool('-l {0} {1}'.format(defm_len, TMP_BIN), 4)
        for line in lines[1:4]:
            self.assertEqual(line[7:], defm)

    def test_option_m(self):
        bin_len = 25
        self._write_bin([0] * bin_len)
        org = 65536 - bin_len
        self._write_ctl(['b {0}'.format(org)])
        defb_mod = 4
        first_defb_len = 4 - (org % defb_mod)
        first_defb = 'DEFB {0}'.format(','.join([str(b) for b in [0] * first_defb_len]))
        defb_len = 8
        defb = 'DEFB {0}'.format(','.join([str(b) for b in [0] * defb_len]))
        lines = self._write_skool('-m {0} {1}'.format(defb_mod, TMP_BIN), 4)
        self.assertEqual(lines[1], 'b{0} {1}'.format(org, first_defb))
        self.assertEqual(lines[2], ' {0} {1}'.format(org + first_defb_len, defb))
        self.assertEqual(lines[3], ' {0} {1}'.format(org + first_defb_len + defb_len, defb))

    def test_option_n(self):
        bin_len = 10
        self._write_bin([0] * bin_len)
        self._write_ctl(['b {0}'.format(65536 - bin_len)])
        defb_len = 3
        defb = 'DEFB {0}'.format(','.join([str(b) for b in [0] * defb_len]))
        lines = self._write_skool('-n {0} {1}'.format(defb_len, TMP_BIN), 4)
        for line in lines[1:4]:
            self.assertEqual(line[7:], defb)

    def test_option_o(self):
        data = [33, 0, 0, 201]
        self._write_bin(data)
        org = str(65536 - len(data) - 100)
        lines = self._write_skool('-o {0} {1}'.format(org, TMP_BIN), 2)
        self.assertEqual(self._get_first_address(lines), org)

    def test_option_r(self):
        data = [201]           # 65533 RET
        data.extend((24, 253)) # 65534 JR 65533
        self._write_bin(data)
        self._write_ctl(['c 65533', 'c 65534'])
        lines = self._write_skool('-r {0}'.format(TMP_BIN), 2)
        self.assertEqual(lines[0], '; Routine at 65533')
        self.assertEqual(lines[1][:10], 'c65533 RET')

    def test_option_s(self):
        start = '24832'
        lines = self._write_skool('-s {0} {1}'.format(start, TEST_BIN), 10)
        self.assertTrue(self._get_first_address(lines) == start)

    def test_option_t(self):
        skool = self._write_skool('-t {0}'.format(TEST_BIN), 10)
        self.assertTrue(' 16385 LD SP,12854   ; [162]' in skool)

    def test_option_z(self):
        data = [0, 1, 2, 3, 4, 5]
        self._write_bin(data)
        org = 65536 - len(data)
        self._write_ctl(['b {0}'.format(org)])
        defb = 'DEFB {0}'.format(','.join(['%03i' % b for b in data]))
        lines = self._write_skool('-z {0}'.format(TMP_BIN), 2)
        self.assertEqual(lines[1], 'b{0} {1}'.format(org, defb))

    def test_sna(self):
        data = [0] * 27 # header
        data += [1, 2, 3] # 16384 LD BC,770
        data += [0] * 49149
        self._write_data(TMP_SNA, data)
        lines = self._write_skool(TMP_SNA, 2)
        self.assertEqual(lines[1][:16], 'c16384 LD BC,770')

    def _write_skool(self, args, split=0):
        output, error = self.run_sna2skool(args, out_lines=False)
        self.assertEqual(len(error), 0)
        if split:
            return [line.rstrip() for line in output.split('\n', split)]
        return output

    def _get_first_address(self, lines):
        for line in lines:
            if line.startswith('c'):
                return line[1:6]

if __name__ == '__main__':
    unittest.main()
